import copy
import pytest
from hpcflow.api import (
    ElementPropagation,
    ValueSequence,
    hpcflow,
    InputSourceType,
    Parameter,
    SchemaInput,
    SchemaOutput,
    TaskSchema,
    TaskObjective,
    TaskSourceType,
    Task,
    InputValue,
    InputSource,
    InputSourceMode,
    Workflow,
    WorkflowTemplate,
)
from hpcflow.sdk.core.errors import (
    MissingInputs,
    TaskTemplateInvalidNesting,
    TaskTemplateMultipleInputValues,
    TaskTemplateMultipleSchemaObjectives,
    TaskTemplateUnexpectedInput,
)


@pytest.fixture
def null_config(tmp_path):
    hpcflow.load_config(config_dir=tmp_path)


@pytest.fixture
def param_p1():
    return Parameter("p1")


@pytest.fixture
def param_p2():
    return Parameter("p2")


@pytest.fixture
def param_p3():
    return Parameter("p3")


@pytest.fixture
def workflow_w1(null_config, tmp_path, param_p1, param_p2):
    s1 = TaskSchema("t1", actions=[], inputs=[param_p1], outputs=[param_p2])
    s2 = TaskSchema("t2", actions=[], inputs=[param_p2])

    t1 = Task(
        schemas=s1,
        sequences=[ValueSequence("inputs.p1", values=[101, 102], nesting_order=1)],
    )
    t2 = Task(schemas=s2, nesting_order={"inputs.p2": 1})

    wkt = WorkflowTemplate(name="w1", tasks=[t1, t2])
    return Workflow.from_template(wkt, path=tmp_path)


@pytest.fixture
def workflow_w2(null_config, tmp_path, param_p1, param_p2):
    s1 = TaskSchema("t1", actions=[], inputs=[param_p1], outputs=[param_p2])
    s2 = TaskSchema("t2", actions=[], inputs=[param_p2, param_p3])

    t1 = Task(
        schemas=s1,
        sequences=[ValueSequence("inputs.p1", values=[101, 102], nesting_order=1)],
    )
    t2 = Task(
        schemas=s2,
        sequences=[ValueSequence("inputs.p3", values=[301, 302, 303], nesting_order=1)],
        nesting_order={"inputs.p2": 0},
    )

    wkt = WorkflowTemplate(name="w1", tasks=[t1, t2])
    return Workflow.from_template(wkt, path=tmp_path)


@pytest.fixture
def workflow_w3(null_config, tmp_path, param_p1, param_p2, param_p3, param_p4):
    s1 = TaskSchema("t1", actions=[], inputs=[param_p1], outputs=[param_p3])
    s2 = TaskSchema("t2", actions=[], inputs=[param_p2, param_p3], outputs=[param_p4])
    s3 = TaskSchema("t3", actions=[], inputs=[param_p3, param_p4])

    t1 = Task(schemas=s1, inputs=[InputValue(param_p1, 101)])
    t2 = Task(
        schemas=s2,
        sequences=[ValueSequence("inputs.p2", values=[201, 202], nesting_order=1)],
    )
    t3 = Task(schemas=s3, nesting_order={"inputs.p3": 0, "inputs.p4": 1})

    wkt = WorkflowTemplate(name="w1", tasks=[t1, t2, t3])
    return Workflow.from_template(wkt, name=wkt.name, overwrite=True)


@pytest.fixture
def workflow_w4(null_config, tmp_path, param_p1, param_p2):
    s1 = TaskSchema("t1", actions=[], inputs=[param_p1], outputs=[param_p2])
    t1 = Task(schemas=s1, inputs=[InputValue(param_p1, 101)])
    wkt = WorkflowTemplate(name="w1", tasks=[t1])
    return Workflow.from_template(wkt, path=tmp_path)


def test_task_expected_input_source_mode_no_sources(param_p1):
    s1 = TaskSchema("t1", actions=[], inputs=[SchemaInput(param_p1)])
    t1 = Task(
        schemas=s1,
        inputs=[InputValue(param_p1, value=101)],
    )
    assert t1.element_sets[0].input_source_mode == InputSourceMode.AUTO


def test_task_expected_input_source_mode_with_sources(param_p1):
    s1 = TaskSchema("t1", actions=[], inputs=[SchemaInput(param_p1)])
    t1 = Task(
        schemas=s1,
        inputs=[InputValue(param_p1, value=101)],
        input_sources=[InputSource.local()],
    )
    assert t1.element_sets[0].input_source_mode == InputSourceMode.MANUAL


def test_task_get_available_task_input_sources_expected_return_first_task_local_value(
    param_p1,
):

    s1 = TaskSchema("t1", actions=[], inputs=[SchemaInput(param_p1)])
    t1 = Task(schemas=s1, inputs=[InputValue(param_p1, value=101)])

    available = t1.get_available_task_input_sources(
        element_set=t1.element_sets[0],
        source_tasks=[],
    )
    available_exp = {"p1": [InputSource(source_type=InputSourceType.LOCAL)]}

    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_first_task_default_value(
    param_p1,
):

    s1 = TaskSchema("t1", actions=[], inputs=[SchemaInput(param_p1, default_value=101)])

    t1 = Task(schemas=s1)

    available = t1.get_available_task_input_sources(element_set=t1.element_sets[0])
    available_exp = {"p1": [InputSource(source_type=InputSourceType.DEFAULT)]}

    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_one_param_one_output(
    param_p1, param_p2
):

    s1 = TaskSchema(
        "t1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2)],
    )
    s2 = TaskSchema("t2", actions=[], inputs=[SchemaInput(param_p2)])

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s2)

    t1._insert_ID = 0

    available = t2.get_available_task_input_sources(
        element_set=t2.element_sets[0],
        source_tasks=[t1],
    )
    available_exp = {
        "p2": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            )
        ]
    }
    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_one_param_one_output_with_default(
    param_p1, param_p2
):

    s1 = TaskSchema(
        "t1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2)],
    )
    s2 = TaskSchema("t2", actions=[], inputs=[SchemaInput(param_p2, default_value=2002)])

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s2)

    t1._insert_ID = 0

    available = t2.get_available_task_input_sources(
        element_set=t2.element_sets[0],
        source_tasks=[t1],
    )
    available_exp = {
        "p2": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            ),
            InputSource(source_type=InputSourceType.DEFAULT),
        ]
    }
    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_one_param_one_output_with_local(
    param_p1, param_p2
):

    s1 = TaskSchema(
        "t1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2)],
    )
    s2 = TaskSchema("t2", actions=[], inputs=[SchemaInput(param_p2)])

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s2, inputs=[InputValue(param_p2, value=202)])

    t1._insert_ID = 0

    available = t2.get_available_task_input_sources(
        element_set=t2.element_sets[0],
        source_tasks=[t1],
    )
    available_exp = {
        "p2": [
            InputSource(source_type=InputSourceType.LOCAL),
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            ),
        ]
    }
    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_one_param_one_output_with_default_and_local(
    param_p1, param_p2
):

    s1 = TaskSchema(
        "t1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2)],
    )
    s2 = TaskSchema("t2", actions=[], inputs=[SchemaInput(param_p2, default_value=2002)])

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s2, inputs=[InputValue(param_p2, value=202)])

    t1._insert_ID = 0

    available = t2.get_available_task_input_sources(
        element_set=t2.element_sets[0],
        source_tasks=[t1],
    )
    available_exp = {
        "p2": [
            InputSource(source_type=InputSourceType.LOCAL),
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            ),
            InputSource(source_type=InputSourceType.DEFAULT),
        ]
    }
    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_one_param_two_outputs(
    param_p1, param_p2, param_p3
):
    s1 = TaskSchema(
        "t1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2), SchemaOutput(param_p3)],
    )
    s2 = TaskSchema(
        "t2",
        actions=[],
        inputs=[SchemaInput(param_p2)],
        outputs=[SchemaOutput(param_p3), SchemaOutput("p4")],
    )
    s3 = TaskSchema("ts3", actions=[], inputs=[SchemaInput(param_p3)])

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s2)
    t3 = Task(schemas=s3)

    t1._insert_ID = 0
    t2._insert_ID = 1

    available = t3.get_available_task_input_sources(
        element_set=t3.element_sets[0],
        source_tasks=[t1, t2],
    )
    available_exp = {
        "p3": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            ),
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=1,
                task_source_type=TaskSourceType.OUTPUT,
            ),
        ]
    }
    assert available == available_exp


def test_task_get_available_task_input_sources_expected_return_two_params_one_output(
    param_p1,
    param_p2,
    param_p3,
):

    s1 = TaskSchema(
        "t1",
        actions=[],
        inputs=[SchemaInput(param_p1)],
        outputs=[SchemaOutput(param_p2), SchemaOutput(param_p3)],
    )
    s2 = TaskSchema(
        "t2",
        actions=[],
        inputs=[SchemaInput(param_p2), SchemaInput(param_p3)],
    )

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s2)

    t1._insert_ID = 0

    available = t2.get_available_task_input_sources(
        element_set=t2.element_sets[0],
        source_tasks=[t1],
    )
    available_exp = {
        "p2": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            )
        ],
        "p3": [
            InputSource(
                source_type=InputSourceType.TASK,
                task_ref=0,
                task_source_type=TaskSourceType.OUTPUT,
            )
        ],
    }
    assert available == available_exp


def test_get_task_unique_names_two_tasks_no_repeats():
    s1 = TaskSchema("t1", actions=[])
    s2 = TaskSchema("t2", actions=[])

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s2)

    assert Task.get_task_unique_names([t1, t2]) == ["t1", "t2"]


def test_get_task_unique_names_two_tasks_with_repeat():

    s1 = TaskSchema("t1", actions=[])

    t1 = Task(schemas=s1)
    t2 = Task(schemas=s1)

    assert Task.get_task_unique_names([t1, t2]) == ["t1_1", "t1_2"]


def test_raise_on_multiple_schema_objectives():

    s1 = TaskSchema("t1", actions=[])
    s2 = TaskSchema("t2", actions=[])
    with pytest.raises(TaskTemplateMultipleSchemaObjectives):
        Task(schemas=[s1, s2])


def test_raise_on_unexpected_inputs(param_p1, param_p2):

    s1 = TaskSchema("t1", actions=[], inputs=[param_p1])

    with pytest.raises(TaskTemplateUnexpectedInput):
        Task(
            schemas=s1,
            inputs=[
                InputValue(param_p1, value=101),
                InputValue(param_p2, value=4),
            ],
        )


def test_raise_on_multiple_input_values(param_p1):

    s1 = TaskSchema("s1", inputs=[param_p1], actions=[])

    with pytest.raises(TaskTemplateMultipleInputValues):
        Task(
            schemas=s1,
            inputs=[
                InputValue(param_p1, value=101),
                InputValue(param_p1, value=7),
            ],
        )


def test_expected_return_defined_and_undefined_input_types(param_p1, param_p2):

    s1 = TaskSchema("s1", inputs=[param_p1, param_p2], actions=[])
    t1 = Task(schemas=s1, inputs=[InputValue(param_p1, value=101)])
    element_set = t1.element_sets[0]
    assert element_set.defined_input_types == {
        param_p1.typ
    } and element_set.undefined_input_types == {param_p2.typ}


def test_expected_return_all_schema_input_types_single_schema(param_p1, param_p2):

    s1 = TaskSchema("t1", actions=[], inputs=[param_p1, param_p2])
    t1 = Task(schemas=s1)

    assert t1.all_schema_input_types == {param_p1.typ, param_p2.typ}


def test_expected_return_all_schema_input_types_multiple_schemas(
    param_p1, param_p2, param_p3
):

    s1 = TaskSchema("t1", actions=[], inputs=[param_p1, param_p2])
    s2 = TaskSchema("t1", actions=[], inputs=[param_p1, param_p3])

    t1 = Task(schemas=[s1, s2])

    assert t1.all_schema_input_types == {param_p1.typ, param_p2.typ, param_p3.typ}


def test_expected_name_single_schema():
    s1 = TaskSchema("t1", actions=[])
    t1 = Task(schemas=[s1])
    assert t1.name == "t1"


def test_expected_name_single_schema_with_method():
    s1 = TaskSchema("t1", method="m1", actions=[])
    t1 = Task(schemas=s1)
    assert t1.name == "t1_m1"


def test_expected_name_single_schema_with_implementation():
    s1 = TaskSchema("t1", implementation="i1", actions=[])
    t1 = Task(schemas=s1)
    assert t1.name == "t1_i1"


def test_expected_name_single_schema_with_method_and_implementation():
    s1 = TaskSchema("t1", method="m1", implementation="i1", actions=[])
    t1 = Task(schemas=s1)
    assert t1.name == "t1_m1_i1"


def test_expected_name_multiple_schemas():
    s1 = TaskSchema("t1", actions=[])
    s2 = TaskSchema("t1", actions=[])
    t1 = Task(schemas=[s1, s2])
    assert t1.name == "t1"


def test_expected_name_two_schemas_first_with_method():
    s1 = TaskSchema("t1", method="m1", actions=[])
    s2 = TaskSchema("t1", actions=[])
    t1 = Task(schemas=[s1, s2])
    assert t1.name == "t1_m1"


def test_expected_name_two_schemas_first_with_method_and_implementation():
    s1 = TaskSchema("t1", method="m1", implementation="i1", actions=[])
    s2 = TaskSchema("t1", actions=[])
    t1 = Task(schemas=[s1, s2])
    assert t1.name == "t1_m1_i1"


def test_expected_name_two_schemas_both_with_method():
    s1 = TaskSchema("t1", method="m1", actions=[])
    s2 = TaskSchema("t1", method="m2", actions=[])
    t1 = Task(schemas=[s1, s2])
    assert t1.name == "t1_m1_and_m2"


def test_expected_name_two_schemas_first_with_method_second_with_implementation():
    s1 = TaskSchema("t1", method="m1", actions=[])
    s2 = TaskSchema("t1", implementation="i2", actions=[])
    t1 = Task(schemas=[s1, s2])
    assert t1.name == "t1_m1_and_i2"


def test_expected_name_two_schemas_first_with_implementation_second_with_method():
    s1 = TaskSchema("t1", implementation="i1", actions=[])
    s2 = TaskSchema("t1", method="m2", actions=[])
    t1 = Task(schemas=[s1, s2])
    assert t1.name == "t1_i1_and_m2"


def test_expected_name_two_schemas_both_with_method_and_implementation():
    s1 = TaskSchema("t1", method="m1", implementation="i1", actions=[])
    s2 = TaskSchema("t1", method="m2", implementation="i2", actions=[])
    t1 = Task(schemas=[s1, s2])
    assert t1.name == "t1_m1_i1_and_m2_i2"


def test_raise_on_negative_nesting_order(param_p1):
    s1 = TaskSchema("t1", inputs=[param_p1], actions=[])
    with pytest.raises(TaskTemplateInvalidNesting):
        Task(schemas=s1, nesting_order={"p1": -1})


# TODO: test resolution of elements and with raise MissingInputs


def test_empty_task_init(param_p1):
    """Check we can init a Task with no input values."""
    s1 = TaskSchema("t1", inputs=[param_p1], actions=[])
    t1 = Task(schemas=s1)


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_task_dependencies(workflow_w1):
    assert workflow_w1.tasks.t2.task_dependencies == [workflow_w1.tasks.t1]


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_dependent_tasks(workflow_w1):
    assert workflow_w1.tasks.t1.dependent_tasks == [workflow_w1.tasks.t2]


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_element_dependencies(workflow_w1):
    assert workflow_w1.tasks.t2.element_dependencies == [0, 1]


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_dependent_elements(workflow_w1):
    assert workflow_w1.tasks.t1.dependent_elements == [2, 3]


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_without_propagation_expected_workflow_num_elements(
    workflow_w1, param_p1
):
    num_elems = workflow_w1.num_elements
    workflow_w1.tasks.t1.add_elements(inputs=[InputValue(param_p1, 103)])
    num_elems_new = workflow_w1.num_elements
    assert num_elems_new - num_elems == 1


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_without_propagation_expected_task_num_elements(
    workflow_w1, param_p1
):
    num_elems = workflow_w1.tasks.t1.num_elements
    workflow_w1.tasks.t1.add_elements(inputs=[InputValue(param_p1, 103)])
    num_elems_new = workflow_w1.tasks.t1.num_elements
    assert num_elems_new - num_elems == 1


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_without_propagation_expected_new_data_index(
    workflow_w1, param_p1
):
    data_index = [sorted(i.data_index.keys()) for i in workflow_w1.elements]
    workflow_w1.tasks.t1.add_elements(inputs=[InputValue(param_p1, 103)])
    data_index_new = [sorted(i.data_index.keys()) for i in workflow_w1.elements]
    new_elems = data_index_new[len(data_index) :]
    assert new_elems == [["inputs.p1", "outputs.p2", "resources.any"]]


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_with_propagation_expected_workflow_num_elements(
    workflow_w1, param_p1
):
    num_elems = workflow_w1.num_elements
    workflow_w1.tasks.t1.add_elements(
        inputs=[InputValue(param_p1, 103)],
        propagate_to=[ElementPropagation(task=workflow_w1.tasks.t2)],
    )
    num_elems_new = workflow_w1.num_elements
    assert num_elems_new - num_elems == 2


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_with_propagation_expected_task_num_elements(
    workflow_w1, param_p1
):
    num_elems = [task.num_elements for task in workflow_w1.tasks]
    workflow_w1.tasks.t1.add_elements(
        inputs=[InputValue(param_p1, 103)],
        propagate_to=[ElementPropagation(task=workflow_w1.tasks.t2)],
    )
    num_elems_new = [task.num_elements for task in workflow_w1.tasks]
    num_elems_diff = [i - j for i, j in zip(num_elems_new, num_elems)]
    assert num_elems_diff[0] == 1 and num_elems_diff[1] == 1


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_with_propagation_expected_new_data_index(
    workflow_w1, param_p1
):
    data_index = [sorted(i.data_index.keys()) for i in workflow_w1.elements]
    workflow_w1.tasks.t1.add_elements(
        inputs=[InputValue(param_p1, 103)],
        propagate_to=[ElementPropagation(task=workflow_w1.tasks.t2)],
    )
    data_index_new = [sorted(i.data_index.keys()) for i in workflow_w1.elements]
    new_elems = data_index_new[len(data_index) :]
    assert new_elems == [
        ["inputs.p1", "outputs.p2", "resources.any"],
        ["inputs.p2", "resources.any"],
    ]


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_sequence_without_propagation_expected_workflow_num_elements(
    workflow_w1,
):
    num_elems = workflow_w1.num_elements
    workflow_w1.tasks.t1.add_elements(
        sequences=[ValueSequence("inputs.p1", values=[103, 104], nesting_order=1)]
    )
    num_elems_new = workflow_w1.num_elements
    assert num_elems_new - num_elems == 2


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_sequence_without_propagation_expected_task_num_elements(
    workflow_w1,
):
    num_elems = workflow_w1.tasks.t1.num_elements
    workflow_w1.tasks.t1.add_elements(
        sequences=[ValueSequence("inputs.p1", values=[103, 104], nesting_order=1)]
    )
    num_elems_new = workflow_w1.tasks.t1.num_elements
    assert num_elems_new - num_elems == 2


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_sequence_without_propagation_expected_new_data_index(
    workflow_w1,
):
    data_index = [sorted(i.data_index.keys()) for i in workflow_w1.elements]
    workflow_w1.tasks.t1.add_elements(
        sequences=[ValueSequence("inputs.p1", values=[103, 104], nesting_order=1)]
    )
    data_index_new = [sorted(i.data_index.keys()) for i in workflow_w1.elements]
    new_elems = data_index_new[len(data_index) :]
    assert new_elems == [
        ["inputs.p1", "outputs.p2", "resources.any"],
        ["inputs.p1", "outputs.p2", "resources.any"],
    ]


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_sequence_with_propagation_expected_workflow_num_elements(
    workflow_w1,
):
    num_elems = workflow_w1.num_elements
    workflow_w1.tasks.t1.add_elements(
        sequences=[ValueSequence("inputs.p1", values=[103, 104, 105], nesting_order=1)],
        propagate_to=[
            ElementPropagation(task=workflow_w1.tasks.t2, nesting_order={"inputs.p2": 1}),
        ],
    )
    num_elems_new = workflow_w1.num_elements
    assert num_elems_new - num_elems == 6


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_sequence_with_propagation_expected_task_num_elements(
    workflow_w1,
):
    num_elems = [task.num_elements for task in workflow_w1.tasks]
    workflow_w1.tasks.t1.add_elements(
        sequences=[ValueSequence("inputs.p1", values=[103, 104, 105], nesting_order=1)],
        propagate_to=[
            ElementPropagation(task=workflow_w1.tasks.t2, nesting_order={"inputs.p2": 1}),
        ],
    )
    num_elems_new = [task.num_elements for task in workflow_w1.tasks]
    num_elems_diff = [i - j for i, j in zip(num_elems_new, num_elems)]
    assert num_elems_diff[0] == 3 and num_elems_diff[1] == 3


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_sequence_with_propagation_expected_new_data_index(workflow_w1):
    data_index = [sorted(i.data_index.keys()) for i in workflow_w1.elements]
    workflow_w1.tasks.t1.add_elements(
        sequences=[ValueSequence("inputs.p1", values=[103, 104, 105], nesting_order=1)],
        propagate_to=[
            ElementPropagation(task=workflow_w1.tasks.t2, nesting_order={"inputs.p2": 1}),
        ],
    )
    data_index_new = [sorted(i.data_index.keys()) for i in workflow_w1.elements]
    new_elems = data_index_new[len(data_index) :]
    assert new_elems == [
        ["inputs.p1", "outputs.p2", "resources.any"],
        ["inputs.p1", "outputs.p2", "resources.any"],
        ["inputs.p1", "outputs.p2", "resources.any"],
        ["inputs.p2", "resources.any"],
        ["inputs.p2", "resources.any"],
        ["inputs.p2", "resources.any"],
    ]


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_sequence_with_propagation_into_sequence_expected_workflow_num_elements(
    workflow_w2,
):
    num_elems = workflow_w2.num_elements
    workflow_w2.tasks.t1.add_elements(
        sequences=[ValueSequence("inputs.p1", values=[103, 104, 105], nesting_order=1)],
        propagate_to=[
            ElementPropagation(
                task=workflow_w2.tasks.t2, nesting_order={"inputs.p2": 1, "inputs.p3": 2}
            ),
        ],
    )
    num_elems_new = workflow_w2.num_elements
    assert num_elems_new - num_elems == 12


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_sequence_with_propagation_into_sequence_expected_task_num_elements(
    workflow_w2,
):
    num_elems = [task.num_elements for task in workflow_w2.tasks]
    workflow_w2.tasks.t1.add_elements(
        sequences=[ValueSequence("inputs.p1", values=[103, 104, 105], nesting_order=1)],
        propagate_to=[
            ElementPropagation(
                task=workflow_w2.tasks.t2, nesting_order={"inputs.p2": 1, "inputs.p3": 2}
            ),
        ],
    )
    num_elems_new = [task.num_elements for task in workflow_w2.tasks]
    num_elems_diff = [i - j for i, j in zip(num_elems_new, num_elems)]
    assert num_elems_diff[0] == 3 and num_elems_diff[1] == 9


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_sequence_with_propagation_into_sequence_expected_new_data_index(
    workflow_w2,
):
    data_index = [sorted(i.data_index.keys()) for i in workflow_w2.elements]
    workflow_w2.tasks.t1.add_elements(
        sequences=[ValueSequence("inputs.p1", values=[103, 104, 105], nesting_order=1)],
        propagate_to=[
            ElementPropagation(
                task=workflow_w2.tasks.t2, nesting_order={"inputs.p2": 1, "inputs.p3": 2}
            ),
        ],
    )
    data_index_new = [sorted(i.data_index.keys()) for i in workflow_w2.elements]
    new_elems = data_index_new[len(data_index) :]
    assert new_elems == [
        ["inputs.p1", "outputs.p2", "resources.any"],
        ["inputs.p1", "outputs.p2", "resources.any"],
        ["inputs.p1", "outputs.p2", "resources.any"],
        ["inputs.p2", "inputs.p3", "resources.any"],
        ["inputs.p2", "inputs.p3", "resources.any"],
        ["inputs.p2", "inputs.p3", "resources.any"],
        ["inputs.p2", "inputs.p3", "resources.any"],
        ["inputs.p2", "inputs.p3", "resources.any"],
        ["inputs.p2", "inputs.p3", "resources.any"],
        ["inputs.p2", "inputs.p3", "resources.any"],
        ["inputs.p2", "inputs.p3", "resources.any"],
        ["inputs.p2", "inputs.p3", "resources.any"],
    ]


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_with_default_propagation(workflow_w1):
    workflow_w1_copy = workflow_w1.copy()

    workflow_w1.tasks.t1.add_elements(
        inputs=[InputValue(param_p1, 103)],
        propagate_to=[],
    )

    workflow_w1_copy.tasks.t1.add_elements(
        inputs=[InputValue(param_p1, 103)],
        propagate_to=[ElementPropagation(task=workflow_w1.tasks.t2)],
    )

    assert workflow_w1._persistent_metadata == workflow_w1_copy._persistent_metadata


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_multi_task_dependence_expected_workflow_num_elements(
    workflow_w3, param_p1
):
    num_elems = workflow_w3.num_elements
    workflow_w3.tasks.t1.add_elements(
        inputs=[InputValue(param_p1, 102)],
        propagate_to=[
            ElementPropagation(
                task=workflow_w3.tasks.t2, nesting_order={"inputs.p2": 0, "inputs.p3": 1}
            ),
            ElementPropagation(
                task=workflow_w3.tasks.t3,
                nesting_order={"inputs.p3": 0, "inputs.p4": 1},
            ),
        ],
    )
    num_elems_new = workflow_w3.num_elements
    assert num_elems_new - num_elems == 5


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_multi_task_dependence_expected_task_num_elements(
    workflow_w3, param_p1
):
    num_elems = [task.num_elements for task in workflow_w3.tasks]
    workflow_w3.tasks.t1.add_elements(
        inputs=[InputValue(param_p1, 102)],
        propagate_to=[
            ElementPropagation(
                task=workflow_w3.tasks.t2, nesting_order={"inputs.p2": 0, "inputs.p3": 1}
            ),
            ElementPropagation(
                task=workflow_w3.tasks.t3,
                nesting_order={"inputs.p3": 0, "inputs.p4": 1},
            ),
        ],
    )
    num_elems_new = [task.num_elements for task in workflow_w3.tasks]
    num_elems_diff = [i - j for i, j in zip(num_elems_new, num_elems)]
    assert num_elems_diff == [1, 2, 2]


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_multi_task_dependence_expected_new_data_index(
    workflow_w3, param_p1
):
    data_index = [sorted(i.data_index.keys()) for i in workflow_w3.elements]
    workflow_w3.tasks.t1.add_elements(
        inputs=[InputValue(param_p1, 102)],
        propagate_to=[
            ElementPropagation(
                task=workflow_w3.tasks.t2, nesting_order={"inputs.p2": 0, "inputs.p3": 1}
            ),
            ElementPropagation(
                task=workflow_w3.tasks.t3,
                nesting_order={"inputs.p3": 0, "inputs.p4": 1},
            ),
        ],
    )
    data_index_new = [sorted(i.data_index.keys()) for i in workflow_w3.elements]
    new_elems = data_index_new[len(data_index) :]
    assert new_elems == (
        [["inputs.p1", "outputs.p3", "resources.any"]] * 1
        + [["inputs.p2", "inputs.p3", "outputs.p4", "resources.any"]] * 2
        + [["inputs.p3", "inputs.p4", "resources.any"]] * 2
    )


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_sequence_multi_task_dependence_workflow_num_elements(
    workflow_w3,
):
    num_elems = workflow_w3.num_elements
    workflow_w3.tasks.t1.add_elements(
        sequences=[ValueSequence("inputs.p1", values=[102, 103, 104], nesting_order=1)],
        propagate_to=[
            ElementPropagation(
                task=workflow_w3.tasks.t2, nesting_order={"inputs.p2": 0, "inputs.p3": 1}
            ),
            ElementPropagation(
                task=workflow_w3.tasks.t3,
                nesting_order={"inputs.p3": 0, "inputs.p4": 1},
            ),
        ],
    )
    num_elems_new = workflow_w3.num_elements
    assert num_elems_new - num_elems == 27


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_sequence_multi_task_dependence_expected_task_num_elements(
    workflow_w3,
):
    num_elems = [task.num_elements for task in workflow_w3.tasks]
    workflow_w3.tasks.t1.add_elements(
        sequences=[ValueSequence("inputs.p1", values=[102, 103, 104], nesting_order=1)],
        propagate_to=[
            ElementPropagation(
                task=workflow_w3.tasks.t2, nesting_order={"inputs.p2": 0, "inputs.p3": 1}
            ),
            ElementPropagation(
                task=workflow_w3.tasks.t3,
                nesting_order={"inputs.p3": 0, "inputs.p4": 1},
            ),
        ],
    )
    num_elems_new = [task.num_elements for task in workflow_w3.tasks]
    num_elems_diff = [i - j for i, j in zip(num_elems_new, num_elems)]
    assert num_elems_diff == [3, 6, 18]


@pytest.mark.skip(
    reason=(
        "Need to be able to either add app data to the app here, or have support for "
        "built in app data; can't init ValueSequence."
    )
)
def test_task_add_elements_sequence_multi_task_dependence_expected_new_data_index(
    workflow_w3,
):
    data_index = [sorted(i.data_index.keys()) for i in workflow_w3.elements]
    workflow_w3.tasks.t1.add_elements(
        sequences=[ValueSequence("inputs.p1", values=[102, 103, 104], nesting_order=1)],
        propagate_to=[
            ElementPropagation(
                task=workflow_w3.tasks.t2, nesting_order={"inputs.p2": 0, "inputs.p3": 1}
            ),
            ElementPropagation(
                task=workflow_w3.tasks.t3,
                nesting_order={"inputs.p3": 0, "inputs.p4": 1},
            ),
        ],
    )
    data_index_new = [sorted(i.data_index.keys()) for i in workflow_w3.elements]
    new_elems = data_index_new[len(data_index) :]
    assert new_elems == (
        [["inputs.p1", "outputs.p3", "resources.any"]] * 3
        + [["inputs.p2", "inputs.p3", "outputs.p4", "resources.any"]] * 6
        + [["inputs.p3", "inputs.p4", "resources.any"]] * 18
    )


def test_no_change_to_persistent_metadata_on_add_task_failure(
    workflow_w4, param_p2, param_p3
):

    data = copy.deepcopy(workflow_w4.metadata)

    s2 = TaskSchema("t2", actions=[], inputs=[param_p2, param_p3])
    t2 = Task(schemas=s2)
    with pytest.raises(MissingInputs) as exc_info:
        workflow_w4.add_task(t2)

    assert workflow_w4.metadata == data


def test_no_change_to_parameter_keys_on_add_task_failure(workflow_w4, param_p2, param_p3):

    param_keys = list(workflow_w4._get_workflow_parameter_group().keys())

    s2 = TaskSchema("t2", actions=[], inputs=[param_p2, param_p3])
    t2 = Task(schemas=s2)
    with pytest.raises(MissingInputs) as exc_info:
        workflow_w4.add_task(t2)

    param_keys_new = list(workflow_w4._get_workflow_parameter_group().keys())

    assert param_keys_new == param_keys


def test_expected_additional_parameter_keys_on_add_task(workflow_w4, param_p2, param_p3):

    param_keys = workflow_w4._get_parameter_keys()
    param_key_max = max(param_keys)

    s2 = TaskSchema("t2", actions=[], inputs=[param_p2, param_p3])
    t2 = Task(schemas=s2, inputs=[InputValue(param_p3, 301)])
    workflow_w4.add_task(t2)

    param_keys_new = workflow_w4._get_parameter_keys()
    param_keys_diff = set(param_keys_new) - set(param_keys)

    # one new key for resources, one for param_p3 value
    assert param_keys_diff == {param_key_max + 1, param_key_max + 2}


def test_parameters_accepted_on_add_task(workflow_w4, param_p2, param_p3):
    s2 = TaskSchema("t2", actions=[], inputs=[param_p2, param_p3])
    t2 = Task(schemas=s2, inputs=[InputValue(param_p3, 301)])
    workflow_w4.add_task(t2)
    assert not workflow_w4._get_pending_add_parameter_keys()


def test_parameters_pending_during_add_task(workflow_w4, param_p2, param_p3):
    s2 = TaskSchema("t2", actions=[], inputs=[param_p2, param_p3])
    t2 = Task(schemas=s2, inputs=[InputValue(param_p3, 301)])
    with workflow_w4.batch_update():
        workflow_w4.add_task(t2)
        assert workflow_w4._get_pending_add_parameter_keys()



from __future__ import annotations
import copy
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Union

from .json_like import ChildObjectSpec, JSONLike
from .command_files import FileSpec
from .element import ElementFilter, ElementGroup
from .errors import (
    MissingInputs,
    TaskTemplateInvalidNesting,
    TaskTemplateMultipleInputValues,
    TaskTemplateMultipleSchemaObjectives,
    TaskTemplateUnexpectedInput,
    TaskTemplateUnexpectedSequenceInput,
)
from .parameters import (
    InputSource,
    InputSourceMode,
    InputSourceType,
    InputValue,
    ParameterPath,
    SchemaInput,
    SchemaOutput,
    ValuePerturbation,
    ValueSequence,
)
from .utils import get_duplicate_items, get_item_repeat_index


class ElementSet(JSONLike):
    """Class to represent a parametrisation of a new set of elements."""

    _child_objects = (
        ChildObjectSpec(
            name="inputs",
            class_name="InputValue",
            is_multiple=True,
            dict_key_attr="parameter",
            dict_val_attr="value",
            parent_ref="_element_set",
        ),
        ChildObjectSpec(
            name="resources",
            class_name="ResourceList",
            parent_ref="_element_set",
        ),
        ChildObjectSpec(
            name="sequences",
            class_name="ValueSequence",
            is_multiple=True,
            parent_ref="_element_set",
        ),
        ChildObjectSpec(
            name="input_sources",
            class_name="InputSource",
            is_multiple=True,
            is_dict_values=True,
            is_dict_values_ensure_list=True,
        ),
        ChildObjectSpec(
            name="input_source_mode",
            class_name="InputSourceMode",
            is_enum=True,
        ),
    )

    def __init__(
        self,
        inputs: Optional[List[InputValue]] = None,
        sequences: Optional[List[ValueSequence]] = None,
        resources: Optional[Dict[str, Dict]] = None,
        repeats: Optional[Union[int, List[int]]] = 1,
        input_sources: Optional[Dict[str, InputSource]] = None,
        input_source_mode: Optional[Union[str, InputSourceType]] = None,
        nesting_order: Optional[List] = None,
        sourceable_elements: Optional[List[int]] = None,
    ):

        if isinstance(resources, dict):
            resources = self.app.ResourceList.from_json_like(resources)
        elif isinstance(resources, list):
            resources = self.app.ResourceList(resources)
        elif not resources:
            resources = self.app.ResourceList([self.app.ResourceSpec()])

        self.inputs = inputs or []
        self.repeats = repeats
        self.resources = resources
        self.sequences = sequences or []
        self.input_sources = input_sources or {}
        self.input_source_mode = input_source_mode or (
            InputSourceMode.MANUAL if input_sources else InputSourceMode.AUTO
        )
        self.nesting_order = nesting_order or {}
        self.sourceable_elements = sourceable_elements

        self._validate()
        self._set_parent_refs()

        self._task_template = None  # assigned by parent Task
        self._defined_input_types = None  # assigned on _task_template assignment

    @classmethod
    def _json_like_constructor(cls, json_like):
        """Invoked by `JSONLike.from_json_like` instead of `__init__`."""
        orig_inp = json_like.pop("original_input_sources", None)
        orig_nest = json_like.pop("original_nesting_order", None)
        obj = cls(**json_like)
        obj.original_input_sources = orig_inp
        obj.original_nesting_order = orig_nest
        return obj

    def prepare_persistent_copy(self):
        """Return a copy of self, which will then be made persistent, and save copies of
        attributes that may be changed during integration with the workflow."""
        obj = copy.deepcopy(self)
        obj.original_nesting_order = self.nesting_order
        obj.original_input_sources = self.input_sources
        return obj

    def to_dict(self):
        dct = super().to_dict()
        del dct["_defined_input_types"]
        del dct["_task_template"]
        return dct

    @property
    def task_template(self):
        return self._task_template

    @task_template.setter
    def task_template(self, value):
        self._task_template = value
        self._validate_against_template()

    @property
    def input_types(self):
        return [i.parameter.typ for i in self.inputs]

    def _validate(self):
        dup_params = get_duplicate_items(self.input_types)
        if dup_params:
            raise TaskTemplateMultipleInputValues(
                f"The following parameters are associated with multiple input value "
                f"definitions: {dup_params!r}."
            )

    def _validate_against_template(self):

        unexpected_types = (
            set(self.input_types) - self.task_template.all_schema_input_types
        )
        if unexpected_types:
            raise TaskTemplateUnexpectedInput(
                f"The following input parameters are unexpected: {list(unexpected_types)!r}"
            )

        seq_inp_types = []
        for seq_i in self.sequences:
            inp_type = seq_i.input_type
            if inp_type:
                bad_inp = {inp_type} - self.task_template.all_schema_input_types
                allowed_str = ", ".join(
                    f'"{i}"' for i in self.task_template.all_schema_input_types
                )
                if bad_inp:
                    raise TaskTemplateUnexpectedSequenceInput(
                        f"The input type {inp_type!r} specified in the following sequence"
                        f" path is unexpected: {seq_i.path!r}. Available input types are: "
                        f"{allowed_str}."
                    )
                seq_inp_types.append(inp_type)
            if seq_i.path not in self.nesting_order:
                self.nesting_order.update({seq_i.path: seq_i.nesting_order})

        for k, v in self.nesting_order.items():
            if v < 0:
                raise TaskTemplateInvalidNesting(
                    f"`nesting_order` must be >=0 for all keys, but for key {k!r}, value "
                    f"of {v!r} was specified."
                )

        self._defined_input_types = set(self.input_types + seq_inp_types)

    @classmethod
    def ensure_element_sets(
        cls,
        inputs=None,
        sequences=None,
        resources=None,
        repeats=None,
        input_sources=None,
        input_source_mode=None,
        nesting_order=None,
        element_sets=None,
        sourceable_elements=None,
    ):
        args = (
            inputs,
            sequences,
            resources,
            repeats,
            input_sources,
            input_source_mode,
            nesting_order,
        )
        args_not_none = [i is not None for i in args]

        if any(args_not_none):
            if element_sets is not None:
                raise ValueError(
                    "If providing an `element_set`, no other arguments are allowed."
                )
            else:
                element_sets = [cls(*args, sourceable_elements=sourceable_elements)]
        else:
            if element_sets is None:
                element_sets = [cls(*args, sourceable_elements=sourceable_elements)]

        return element_sets

    @property
    def defined_input_types(self):
        return self._defined_input_types

    @property
    def undefined_input_types(self):
        return self.task_template.all_schema_input_types - self.defined_input_types

    def get_sequence_from_path(self, sequence_path):
        for i in self.sequences:
            if i.path == sequence_path:
                return i

    def get_defined_parameter_types(self):
        out = []
        for inp in self.inputs:
            if not inp.is_sub_value:
                out.append(inp.normalised_inputs_path)
        for seq in self.sequences:
            if seq.parameter and not seq.is_sub_value:  # ignore resource sequences
                out.append(seq.normalised_inputs_path)
        return out

    def get_defined_sub_parameter_types(self):
        out = []
        for inp in self.inputs:
            if inp.is_sub_value:
                out.append(inp.normalised_inputs_path)
        for seq in self.sequences:
            if seq.parameter and seq.is_sub_value:  # ignore resource sequences
                out.append(seq.normalised_inputs_path)
        return out

    def get_locally_defined_inputs(self):
        return self.get_defined_parameter_types() + self.get_defined_sub_parameter_types()

    def get_sequence_by_path(self, path):
        for seq in self.sequences:
            if seq.path == path:
                return seq

    @property
    def index(self):
        for idx, element_set in enumerate(self.task_template.element_sets):
            if element_set is self:
                return idx

    @property
    def elements(self):
        task = self.task_template.workflow_template.workflow.tasks[
            self.task_template.index
        ]
        elements = task.get_elements_of_element_set(self.index)
        return elements

    @property
    def task_dependencies(self):
        """Get tasks that this element set depends on."""

        dependencies = []
        for element in self.elements:
            for task_dep_i in element.task_dependencies:
                if task_dep_i not in dependencies:
                    dependencies.append(task_dep_i)

        return dependencies

    @property
    def dependent_tasks(self):
        """Get tasks that depend on this element set."""

        dependents = []
        for element in self.elements:
            for task_dep_i in element.dependent_tasks:
                if task_dep_i not in dependents:
                    dependents.append(task_dep_i)
        return dependents

    @property
    def element_dependencies(self):
        """Get indices of elements that this element set depends on."""

        dependencies = []
        for element in self.elements:
            for elem_dep_i in element.element_dependencies:
                if elem_dep_i not in dependencies:
                    dependencies.append(elem_dep_i)

        return dependencies

    @property
    def dependent_elements(self):
        """Get indices of elements that depend on this element set."""

        dependents = []
        for element in self.elements:
            for elem_dep_i in element.dependent_elements:
                if elem_dep_i not in dependents:
                    dependents.append(elem_dep_i)
        return dependents


class Task(JSONLike):
    """Parametrisation of an isolated task for which a subset of input values are given
    "locally". The remaining input values are expected to be satisfied by other
    tasks/imports in the workflow."""

    _child_objects = (
        ChildObjectSpec(
            name="schemas",
            class_name="TaskSchema",
            is_multiple=True,
            shared_data_name="task_schemas",
            shared_data_primary_key="name",
            parent_ref="_task_template",
        ),
        ChildObjectSpec(
            name="element_sets",
            class_name="ElementSet",
            is_multiple=True,
            parent_ref="task_template",
        ),
    )

    def __init__(
        self,
        schemas: Union[TaskSchema, str, List[TaskSchema], List[str]],
        repeats: Optional[Union[int, List[int]]] = None,
        resources: Optional[Dict[str, Dict]] = None,
        inputs: Optional[List[InputValue]] = None,
        sequences: Optional[List[ValueSequence]] = None,
        input_sources: Optional[Dict[str, InputSource]] = None,
        input_source_mode: Optional[Union[str, InputSourceType]] = None,
        nesting_order: Optional[List] = None,
        element_sets: Optional[List[ElementSet]] = None,
        sourceable_elements: Optional[List[int]] = None,
    ):

        """
        Parameters
        ----------
        schema
            A (list of) `TaskSchema` object(s) and/or a (list of) strings that are task
            schema names that uniquely identify a task schema. If strings are provided,
            the `TaskSchema` object will be fetched from the known task schemas loaded by
            the app configuration.

        """

        # TODO: allow init via specifying objective and/or method and/or implementation
        # (lists of) strs e.g.: Task(
        #   objective='simulate_VE_loading',
        #   method=['CP_FFT', 'taylor'],
        #   implementation=['damask', 'damask']
        # )
        # where method and impl must be single strings of lists of the same length
        # and method/impl are optional/required only if necessary to disambiguate
        #
        # this would be like Task(schemas=[
        #   'simulate_VE_loading_CP_FFT_damask',
        #   'simulate_VE_loading_taylor_damask'
        # ])

        if not isinstance(schemas, list):
            schemas = [schemas]

        _schemas = []
        for i in schemas:
            if isinstance(i, str):
                try:
                    i = self.app.TaskSchema.get_by_key(i)
                except KeyError:
                    raise KeyError(f"TaskSchema {i!r} not found.")
            elif not isinstance(i, self.app.TaskSchema):
                raise TypeError(f"Not a TaskSchema object: {i!r}")
            _schemas.append(i)

        self._schemas = _schemas

        self._element_sets = self.app.ElementSet.ensure_element_sets(
            inputs=inputs,
            sequences=sequences,
            resources=resources,
            repeats=repeats,
            input_sources=input_sources,
            input_source_mode=input_source_mode,
            nesting_order=nesting_order,
            element_sets=element_sets,
            sourceable_elements=sourceable_elements,
        )

        self._validate()
        self._name = self._get_name()

        self.workflow_template = None  # assigned by parent WorkflowTemplate
        self._insert_ID = None
        self._dir_name = None

        self._set_parent_refs()

    @classmethod
    def _json_like_constructor(cls, json_like):
        """Invoked by `JSONLike.from_json_like` instead of `__init__`."""
        insert_ID = json_like.pop("insert_ID", None)
        dir_name = json_like.pop("dir_name", None)
        obj = cls(**json_like)
        obj._insert_ID = insert_ID
        obj._dir_name = dir_name
        return obj

    def __repr__(self):
        return f"{self.__class__.__name__}(" f"name={self.name!r}" f")"

    def to_persistent(self, workflow):
        """Make a copy where any schema input defaults are saved to a persistent
        workflow. ElementSet data is not made persistent."""

        obj = copy.deepcopy(self)
        new_param_groups = []
        for schema in obj.schemas:
            new_param_groups.extend(schema.make_persistent(workflow))

        return obj, new_param_groups

    def to_dict(self):
        out = super().to_dict()
        return {k.lstrip("_"): v for k, v in out.items() if k != "_name"}

    def _validate(self):

        # TODO: check a nesting order specified for each sequence?

        names = set(i.objective.name for i in self.schemas)
        if len(names) > 1:
            raise TaskTemplateMultipleSchemaObjectives(
                f"All task schemas used within a task must have the same "
                f"objective, but found multiple objectives: {list(names)!r}"
            )

    def _get_name(self):
        out = f"{self.objective.name}"
        for idx, schema_i in enumerate(self.schemas, start=1):
            need_and = idx < len(self.schemas) and (
                self.schemas[idx].method or self.schemas[idx].implementation
            )
            out += (
                f"{f'_{schema_i.method}' if schema_i.method else ''}"
                f"{f'_{schema_i.implementation}' if schema_i.implementation else ''}"
                f"{f'_and' if need_and else ''}"
            )
        return out

    @staticmethod
    def get_task_unique_names(tasks: List[Task]):
        """Get the unique name of each in a list of tasks.

        Returns
        -------
        list of str

        """

        task_name_rep_idx = get_item_repeat_index(
            tasks,
            item_callable=lambda x: x.name,
            distinguish_singular=True,
        )

        names = []
        for idx, task in enumerate(tasks):
            add_rep = f"_{task_name_rep_idx[idx]}" if task_name_rep_idx[idx] > 0 else ""
            names.append(f"{task.name}{add_rep}")

        return names

    def _get_nesting_order(self, seq):
        """Find the nesting order for a task sequence."""
        return self.nesting_order[seq.normalised_path] if len(seq.values) > 1 else -1

    def _prepare_persistent_outputs(self, workflow, num_elements):
        # TODO: check that schema is present when adding task? (should this be here?)
        output_data_indices = {}
        for schema in self.schemas:
            for output in schema.outputs:
                output_data_indices[output.typ] = []
                for _ in range(num_elements):
                    group_idx = workflow._add_parameter_group(
                        data=None,
                        is_pending_add=workflow._in_batch_mode,
                        is_set=False,
                    )
                    output_data_indices[output.typ].append(group_idx)

        return output_data_indices

    def prepare_element_resolution(self, element_set, input_data_indices):

        multiplicities = []
        for path_i, inp_idx_i in input_data_indices.items():
            multiplicities.append(
                {
                    "multiplicity": len(inp_idx_i),
                    "nesting_order": element_set.nesting_order.get(path_i, -1),
                    "path": path_i,
                }
            )

        return multiplicities

    @property
    def index(self):
        if self.workflow_template:
            return self.workflow_template.tasks.index(self)
        else:
            return None

    @property
    def _element_indices(self):
        return self.workflow_template.workflow.tasks[self.index].element_indices

    def get_available_task_input_sources(
        self,
        element_set: ElementSet,
        source_tasks: Optional[List[Task]] = None,
    ) -> List[InputSource]:
        """For each input parameter of this task, generate a list of possible input sources
        that derive from inputs or outputs of this and other provided tasks.

        Note this only produces a subset of available input sources for each input
        parameter; other available input sources may exist from workflow imports."""

        # TODO: also search sub-parameters in the source tasks!

        available = {}
        for inputs_path, def_val in self.get_all_inputs_info(element_set).items():

            available[inputs_path] = []

            # local specification takes precedence:
            if inputs_path in element_set.get_locally_defined_inputs():
                available[inputs_path].append(self.app.InputSource.local())

            # search for task sources:
            for src_task_i in source_tasks or []:

                for param_i in src_task_i.provides_parameters:

                    if param_i.typ == inputs_path:

                        src_elems = None
                        if element_set.sourceable_elements is not None:
                            # can only use a subset of elements:
                            src_elems = list(
                                set(element_set.sourceable_elements)
                                & set(src_task_i._element_indices)
                            )
                            if not src_elems:
                                continue

                        task_source = self.app.InputSource.task(
                            task_ref=src_task_i.insert_ID,
                            task_source_type=param_i.input_or_output,
                            elements=src_elems,
                        )
                        available[inputs_path].append(task_source)

            if def_val:
                available[inputs_path].append(self.app.InputSource.default())

        return available

    @property
    def schemas(self):
        return self._schemas

    @property
    def element_sets(self):
        return self._element_sets

    @property
    def insert_ID(self):
        return self._insert_ID

    @property
    def dir_name(self):
        "Artefact directory name."
        return self._dir_name

    @property
    def name(self):
        return self._name

    @property
    def objective(self):
        return self.schemas[0].objective

    @property
    def all_schema_inputs(self) -> Tuple[SchemaInput]:
        return tuple(inp_j for schema_i in self.schemas for inp_j in schema_i.inputs)

    @property
    def all_schema_outputs(self) -> Tuple[SchemaOutput]:
        return tuple(inp_j for schema_i in self.schemas for inp_j in schema_i.outputs)

    @property
    def all_schema_input_types(self):
        """Get the set of all schema input types (over all specified schemas)."""
        return {inp_j for schema_i in self.schemas for inp_j in schema_i.input_types}

    @property
    def all_schema_input_normalised_paths(self):
        return {f"inputs.{i}" for i in self.all_schema_input_types}

    @property
    def all_schema_output_types(self):
        """Get the set of all schema output types (over all specified schemas)."""
        return {out_j for schema_i in self.schemas for out_j in schema_i.output_types}

    @property
    def all_sourced_normalised_paths(self):
        sourced_input_types = []
        for elem_set in self.element_sets:
            for inp in elem_set.inputs:
                if inp.is_sub_value:
                    sourced_input_types.append(inp.normalised_path)
            for seq in elem_set.sequences:
                if seq.is_sub_value:
                    sourced_input_types.append(seq.normalised_path)
        return set(sourced_input_types) | self.all_schema_input_normalised_paths

    def get_all_inputs_info(self, element_set):
        """Get a dict whose keys are the normalised paths (without the "inputs" prefix),
        and whose values are the associated default value InputValue object, in the case
        the input is a SchemaInput, and a default is defined.

        Parameters
        ----------
        element_set : ElementSet
            Find inputs and sequences in this element set that have sub-parameter paths.

        """

        info = {}
        for schema_input in self.all_schema_inputs:
            info[schema_input.parameter.typ] = schema_input.default_value

        for inp_path in element_set.get_defined_sub_parameter_types():
            info[inp_path] = None

        return info

    @property
    def all_sequences_normalised_paths(self):
        return [j.normalised_path for i in self.element_sets for j in i.sequences]

    @property
    def all_used_sequences_normalised_paths(self):
        return [
            j.normalised_path
            for i in self.element_sets
            for j in i.sequences
            if not j.is_unused
        ]

    @property
    def universal_input_types(self):
        """Get input types that are associated with all schemas"""

    @property
    def non_universal_input_types(self):
        """Get input types for each schema that are non-universal."""

    @property
    def defined_input_types(self):
        return self._defined_input_types

    @property
    def undefined_input_types(self):
        return self.all_schema_input_types - self.defined_input_types

    @property
    def undefined_inputs(self):
        return [
            inp_j
            for schema_i in self.schemas
            for inp_j in schema_i.inputs
            if inp_j.typ in self.undefined_input_types
        ]

    @property
    def unsourced_inputs(self):
        """Get schema input types for which no input sources are currently specified."""
        return self.all_schema_input_types - set(self.input_sources.keys())

    @property
    def provides_parameters(self):
        return tuple(j for schema in self.schemas for j in schema.provides_parameters)

    @property
    def _metadata(self):
        return self.workflow_template.workflow.metadata["template"]["tasks"][self.index]

    def get_sub_parameter_input_values(self):
        return [i for i in self.inputs if i.is_sub_value]

    def get_non_sub_parameter_input_values(self):
        return [i for i in self.inputs if not i.is_sub_value]

    def add_group(
        self, name: str, where: ElementFilter, group_by_distinct: ParameterPath
    ):
        group = ElementGroup(name=name, where=where, group_by_distinct=group_by_distinct)
        self.groups.add_object(group)


class WorkflowTask:
    """Class to represent a Task that is bound to a Workflow."""

    _app_attr = "app"

    def __init__(
        self,
        template: Task,
        element_indices: List,
        element_input_sources: Dict,
        element_set_indices: List,
        element_sequence_indices: Dict,
        index: int,
        workflow: Workflow,
    ):

        self._template = template
        self._element_indices = element_indices
        self._element_input_sources = element_input_sources
        self._element_set_indices = element_set_indices
        self._element_sequence_indices = element_sequence_indices
        self._workflow = workflow
        self._index = index

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.unique_name!r})"

    @property
    def template(self):
        # via the workflow so that we see any metadata updates:
        return self.workflow.tasks[self.index]._template

    @property
    def element_indices(self):
        return self._element_indices

    @property
    def element_input_sources(self):
        return self._element_input_sources

    @property
    def element_set_indices(self):
        return self._element_set_indices

    @property
    def element_sequence_indices(self):
        return self._element_sequence_indices

    @property
    def elements(self):
        return [self.workflow.elements[i] for i in self.element_indices]

    @property
    def workflow(self):
        return self._workflow

    @property
    def num_elements(self):
        return len(self.element_indices)

    @property
    def index(self):
        """Zero-based position within the workflow. Uses initial index if appending to the
        workflow is not complete."""
        return self._index

    @property
    def name(self):
        return self.template.name

    @property
    def insert_ID(self):
        return self.template.insert_ID

    @property
    def dir_name(self):
        return self.template.dir_name

    @property
    def dir_path(self):
        return self.workflow.path / "tasks" / self.dir_name

    @property
    def unique_name(self):
        return self.workflow.get_task_unique_names()[self.index]

    @property
    def element_dir_list_file_path(self):
        return self.dir_path / "element_dirs.txt"

    @property
    def run_script_file_path(self):
        return self.dir_path / "run_script.ps1"

    @property
    def num_element_sets(self):
        return len(
            self.workflow.metadata["template"]["tasks"][self.index]["element_sets"]
        )

    @property
    def _metadata(self):
        return self.workflow.metadata["tasks"][self.index]

    def write_element_dirs(self):
        self.dir_path.mkdir(exist_ok=True, parents=True)
        elem_paths = [self.dir_path / elem.dir_name for elem in self.elements]
        for path_i in elem_paths:
            path_i.mkdir(exist_ok=True)

        # write a text file whose lines correspond to element paths
        with self.element_dir_list_file_path.open("wt") as fp:
            for elem in elem_paths:
                fp.write(f"{elem}\n")

    def _make_new_elements_persistent(self, element_set):

        input_data_indices = {}
        input_sources = {}
        sequence_idx = {}
        new_param_groups = []

        # Assign first assuming all locally defined values are to be used:
        for res_i in element_set.resources:
            key, group, is_new = res_i.make_persistent(self.workflow)
            input_data_indices[key] = group
            new_param_groups.extend(group) if is_new else None

        for inp_i in element_set.inputs:
            key, group, is_new = inp_i.make_persistent(self.workflow)
            input_data_indices[key] = group
            input_sources[key] = ["local" for _ in group]
            new_param_groups.extend(group) if is_new else None

        for seq_i in element_set.sequences:
            key, group, is_new = seq_i.make_persistent(self.workflow)
            input_data_indices[key] = group
            input_sources[key] = ["local" for _ in group]
            sequence_idx[key] = list(range(len(group)))
            new_param_groups.extend(group) if is_new else None

        # Now check for task- and default-sources and overwrite or append to local sources:
        for schema_input in self.template.all_schema_inputs:

            key = f"inputs.{schema_input.typ}"
            sources = element_set.input_sources[schema_input.typ]

            for inp_src in sources:

                if inp_src.source_type is InputSourceType.TASK:

                    src_task = inp_src.get_task(self.workflow)

                    src_elements = [i for i in src_task.elements]
                    if inp_src.elements:
                        src_elements = [
                            i for i in src_elements if i.global_index in inp_src.elements
                        ]

                    if not src_elements:
                        continue

                    grp_idx = [
                        elem.data_index[f"outputs.{schema_input.typ}"]
                        for elem in src_elements
                    ]
                    inp_src_i = [
                        f"element.{i.global_index}.{inp_src.task_source_type.name}"
                        for i in src_elements
                    ]

                    if self.app.InputSource.local() in sources:
                        # add task source to existing local source:
                        input_data_indices[key] += grp_idx
                        input_sources[key] += inp_src_i

                    else:
                        # overwrite existing local source (if it exists):
                        input_data_indices[key] = grp_idx
                        input_sources[key] = inp_src_i
                        if key in sequence_idx:
                            sequence_idx.pop(key)
                            seq = element_set.get_sequence_by_path(key)
                            seq.is_unused = True

                if inp_src.source_type is InputSourceType.DEFAULT:

                    grp_idx = [schema_input.default_value._value_group_idx]
                    if self.app.InputSource.local() in sources:
                        input_data_indices[key] += grp_idx
                        input_sources[key] += ["default"]

                    else:
                        input_data_indices[key] = grp_idx
                        input_sources[key] = ["default"]

        return input_data_indices, input_sources, sequence_idx, new_param_groups

    def ensure_input_sources(self, element_set):
        """Check valid input sources are specified for a new task to be added to the
        workflow in a given position. If none are specified, set them according to the
        default behaviour."""

        # this just depends on this schema and other schemas:
        available_sources = self.template.get_available_task_input_sources(
            element_set=element_set,
            source_tasks=self.workflow.template.tasks[: self.index],
        )  # TODO: test all parameters have a key here?

        # TODO: get available input sources from workflow imports

        all_inputs_info = self.template.get_all_inputs_info(element_set)

        # check any specified sources are valid:
        for inputs_path in all_inputs_info:
            for specified_source in element_set.input_sources.get(inputs_path, []):
                self.workflow._resolve_input_source_task_reference(
                    specified_source, self.unique_name
                )
                if not specified_source.is_in(available_sources[inputs_path]):
                    raise ValueError(
                        f"The input source {specified_source.to_string()!r} is not "
                        f"available for input path {inputs_path!r}. Available "
                        f"input sources are: "
                        f"{[i.to_string() for i in available_sources[inputs_path]]}"
                    )

        # TODO: if an input is not specified at all in the `inputs` dict (what about when list?),
        # then check if there is an input files entry for associated inputs,
        # if there is

        unsourced_inputs = set(all_inputs_info.keys()) - set(
            element_set.input_sources.keys()
        )

        # set source for any unsourced inputs:
        missing = []
        for input_type in unsourced_inputs:
            inp_i_sources = available_sources[input_type]
            source = None
            try:
                # first element is defined by default to take precedence in
                # `get_available_task_input_sources`:
                source = inp_i_sources[0]
            except IndexError:
                missing.append(input_type)

            if source is not None:
                element_set.input_sources.update({input_type: [source]})

        if missing:
            missing_str = ", ".join(f"{i!r}" for i in missing)
            raise MissingInputs(
                message=f"The following inputs have no sources: {missing_str}.",
                missing_inputs=missing,
            )

    def _add_element_set(self, element_set, parent_events):
        """
        Returns
        -------
        element_indices : list of int
            Global indices of newly added elements.

        """

        element_set.task_template = self.template  # may modify element_set.nesting_order
        self.ensure_input_sources(element_set)  # may modify element_set.input_sources

        (
            input_data_idx,
            input_sources,
            seq_idx,
            new_param_groups,
        ) = self._make_new_elements_persistent(element_set)

        multiplicities = self.template.prepare_element_resolution(
            element_set, input_data_idx
        )
        element_data_idx = self.workflow.resolve_element_data_indices(multiplicities)
        output_data_idx = self.template._prepare_persistent_outputs(
            self.workflow, len(element_data_idx)
        )

        (
            new_elements,
            element_input_sources,
            element_seq_idx,
        ) = self.workflow.generate_new_elements(
            input_data_idx,
            output_data_idx,
            element_data_idx,
            input_sources,
            seq_idx,
        )

        element_indices = list(
            range(
                len(self.workflow.elements),
                len(self.workflow.elements) + len(new_elements),
            )
        )

        evt = self.workflow.event_log.event_add_element_set(
            task_index=self.index,
            new_element_indices=element_indices,
            new_parameter_groups=new_param_groups,
            parents=parent_events,
        )

        element_set_js, _ = element_set.to_json_like()
        # (shared data should already have been updated as part of the schema)

        # need to add the element_set first so we can update `element_input_sources` and
        # `element_sequence_indices` keys to include potentially new sequence/input paths
        # from this element set:

        self.template._metadata["element_sets"].append(element_set_js)
        self.workflow._save_metadata()

        self._metadata["element_input_sources"] = {
            **{k: [] for k in self.template.all_sourced_normalised_paths},
            **self._metadata["element_input_sources"],
        }

        self._metadata["element_sequence_indices"] = {
            **{
                k: [None for _ in range(self.num_elements)]
                for k in self.template.all_used_sequences_normalised_paths
            },
            **self._metadata["element_sequence_indices"],
        }

        # Now update the remaining metadata:
        self.workflow.metadata["elements"].extend(new_elements)
        self._metadata["element_indices"].extend(element_indices)

        for k, v in self._metadata["element_input_sources"].items():
            v.extend(element_input_sources.get(k, [None] * len(new_elements)))

        for k, v in self._metadata["element_sequence_indices"].items():
            v.extend(element_seq_idx.get(k, [None] * len(new_elements)))

        self._metadata["element_set_indices"].extend(
            [self.num_element_sets - 1] * len(new_elements)
        )

        self.workflow._save_metadata()

        return element_indices

    @property
    def upstream_tasks(self):
        return [task for task in self.workflow.tasks[: self.index]]

    @property
    def downstream_tasks(self):
        return [task for task in self.workflow.tasks[self.index + 1 :]]

    def get_elements_of_element_set(self, set_index):
        idx = [idx for idx, i in enumerate(self.element_set_indices) if i == set_index]
        elements = [self.workflow.elements[self.element_indices[i]] for i in idx]
        return elements

    def add_elements(
        self,
        base_element=None,
        inputs=None,
        sequences=None,
        resources=None,
        repeats=None,
        input_sources=None,
        input_source_mode=None,
        nesting_order=None,
        element_sets=None,
        sourceable_elements=None,
        propagate_to=None,
        return_indices=False,
    ):
        with self.workflow.batch_update():
            return self._add_elements(
                base_element=base_element,
                inputs=inputs,
                sequences=sequences,
                resources=resources,
                repeats=repeats,
                input_sources=input_sources,
                input_source_mode=input_source_mode,
                nesting_order=nesting_order,
                element_sets=element_sets,
                sourceable_elements=sourceable_elements,
                propagate_to=propagate_to,
                return_indices=return_indices,
                parent_events=[],
            )

    def _add_elements(
        self,
        base_element=None,
        inputs=None,
        sequences=None,
        resources=None,
        repeats=None,
        input_sources=None,
        input_source_mode=None,
        nesting_order=None,
        element_sets=None,
        sourceable_elements=None,
        propagate_to=None,
        return_indices=False,
        parent_events=None,
    ):
        """Add more elements to this task.

        Parameters
        ----------
        sourceable_elements : list of int, optional
            If specified, a list of global element indices from which inputs
            may be sourced. If not specified, all workflow elements are considered
            sourceable.
        propagate_to : list of ElementPropagation, optional
            If specified as an empty or non-empty list, propagate the new elements
            downstream. If an `ElementPropagation` object is not specified for a given
            task, propagation will be attempted using default behaviour.
        return_indices : bool, optional
            If True, return the list of indices of the newly added elements. False by
            default.

        """

        evt = self.workflow.event_log.event_add_elements(
            task_index=self.index,
            parents=parent_events,
        )

        if base_element is not None:
            if base_element.task is not self:
                raise ValueError("If specified, `base_element` must belong to this task.")
            b_inputs, b_resources = base_element.to_element_set_data()
            inputs = inputs or b_inputs
            resources = resources or b_resources

        element_sets = self.app.ElementSet.ensure_element_sets(
            inputs=inputs,
            sequences=sequences,
            resources=resources,
            repeats=repeats,
            input_sources=input_sources,
            input_source_mode=input_source_mode,
            nesting_order=nesting_order,
            element_sets=element_sets,
            sourceable_elements=sourceable_elements,
        )

        parent_events = (parent_events or []) + [evt.index]
        elem_idx = []
        for elem_set_i in element_sets:
            elem_set_i = elem_set_i.prepare_persistent_copy()
            elem_idx += self._add_element_set(elem_set_i, parent_events=parent_events)

        if propagate_to is not None:

            # TODO: also accept a dict as func arg:
            propagate_to = {i.task.unique_name: i for i in propagate_to}

            for task in self.downstream_tasks:

                elem_propagate = propagate_to.get(
                    task.unique_name, ElementPropagation(task=task)
                )
                if self.unique_name not in (
                    i.unique_name for i in elem_propagate.element_set.task_dependencies
                ):
                    # TODO: why can't we just do
                    #  `if self in not elem_propagate.element_set.task_dependencies:`?
                    continue

                # TODO: generate a new ElementSet for this task;
                #       Assume for now we use a single base element set.
                #       Later, allow combining multiple element sets.

                elem_set_i = self.app.ElementSet(
                    inputs=elem_propagate.element_set.inputs,
                    sequences=elem_propagate.element_set.sequences,
                    resources=elem_propagate.element_set.resources,
                    repeats=elem_propagate.element_set.repeats,
                    nesting_order=elem_propagate.nesting_order,
                    sourceable_elements=elem_idx,
                )
                prop_elem_idx = task._add_elements(
                    element_sets=[elem_set_i],
                    return_indices=True,
                    parent_events=parent_events,
                )
                elem_idx.extend(prop_elem_idx)

        if return_indices:
            return elem_idx

    @property
    def task_dependencies(self):
        """Get tasks that this task depends on."""

        dependencies = []
        for element in self.elements:
            for task_dep_i in element.task_dependencies:
                if task_dep_i not in dependencies:
                    dependencies.append(task_dep_i)

        return dependencies

    @property
    def dependent_tasks(self):
        """Get tasks that depend on this task."""
        dependents = []

        # get direct dependents:
        for task in self.downstream_tasks:
            for elem_set in task.template.element_sets:
                for src_lst in elem_set.input_sources.values():
                    for src in src_lst:
                        if src.source_type is self.app.InputSourceType.TASK:
                            if src.task_ref == self.index:
                                if task not in dependents:
                                    dependents.append(task)

        # get indirect dependents:
        for task in dependents:
            for task_i_dep in task.dependent_tasks:
                if task_i_dep not in dependents:
                    dependents.append(task_i_dep)

        return dependents

    @property
    def element_dependencies(self):
        """Get indices of elements that this task depends on."""
        dependencies = []
        for element in self.elements:
            for elem_dep_i in element.element_dependencies:
                if elem_dep_i not in dependencies:
                    dependencies.append(elem_dep_i)

        return dependencies

    @property
    def dependent_elements(self):
        """Get elements that depend on this task."""
        dependents = []
        for element in self.elements:
            for elem_dep_i in element.dependent_elements:
                if elem_dep_i not in dependents:
                    dependents.append(elem_dep_i)

        return dependents


@dataclass
class ElementPropagation:
    """Class to represent how a newly added element set should propagate to a given
    downstream task."""

    task: Task
    element_sets: Optional[Union[List[int], List[ElementSet]]] = None
    nesting_order: Optional[Dict] = None

    @property
    def element_set(self):
        # TEMP property; for now just use the first element set as the base:
        return self.task.template.element_sets[0]



import pytest

from hpcflow.api import TaskSchema, hpcflow, Parameter


def test_shared_data_from_json_like_with_shared_data_dependency():
    """Check we can generate some shared data objects where one depends on another."""

    p1 = Parameter("p1")
    p1._set_hash()
    p1_hash = p1._hash_value

    ts1 = TaskSchema(objective="ts1", actions=[], inputs=[p1])
    ts1._set_hash()
    ts1_hash = ts1._hash_value

    shared_data_json = {
        "parameters": {
            p1_hash: {
                "is_file": p1.is_file,
                "sub_parameters": [],
                "type": p1.typ,
            }
        },
        "task_schemas": {
            ts1_hash: {
                "method": ts1.method,
                "implementation": ts1.implementation,
                "version": ts1.version,
                "objective": ts1.objective.name,
                "inputs": [
                    {
                        "group": None,
                        "where": None,
                        "parameter": f"hash:{p1_hash}",
                        "default_value": None,
                        "propagation_mode": "IMPLICIT",
                    }
                ],
                "outputs": [],
                "actions": [],
            }
        },
    }

    sh = hpcflow.shared_data_from_json_like(shared_data_json)

    assert sh["parameters"] == hpcflow.ParametersList([p1]) and sh[
        "task_schemas"
    ] == hpcflow.TaskSchemasList([ts1])



from __future__ import annotations
import copy
from dataclasses import dataclass, field
from typing import Dict, List
from datetime import datetime

from colorama import init as colorama_init
from termcolor import colored
import numcodecs
import numpy as np

from .utils import get_md5_hash, remove_ansi_escape_sequences


@dataclass
class EventLogItem:

    event_log: EventLog = field(repr=False)
    index: int
    time: np.datetime64
    event_type: str
    machine: str
    content: Dict
    depth: int
    parent: int

    def __post_init__(self):
        if self.parent == -1:
            self.parent = None


class EventLog:
    """Class to track how a workflow was generated."""

    _metadata_dtype = np.dtype(
        {
            "names": [
                "index",
                "time",
                "event_type",
                "machine",
                "depth",
                "parent",
            ],
            "formats": ["int8", "M8[us]", "int8", "int8", "int8", "int8"],
        }
    )

    def __init__(self, workflow):
        colorama_init(autoreset=True)

        self.workflow = workflow

        self._set_original_metadata_lookup()
        self._metadata_lookup = copy.deepcopy(self._original_metadata_lookup)

        disk_mdl = self._get_zarr_group("r").attrs["metadata_lookup"]

        self._new_event_metadata = []
        self._new_event_content = []

    def __eq__(self, other):

        if np.all(self.metadata[:] == other.metadata[:]):
            if np.all(self.content[:] == other.content[:]):
                if self._metadata_lookup == other._metadata_lookup:
                    return True

        return False

    def _set_original_metadata_lookup(self):
        self._original_metadata_lookup = self._get_zarr_group("r").attrs[
            "metadata_lookup"
        ]

    def _check_metadata_lookup_changed(self):
        h1 = get_md5_hash(self._metadata_lookup)
        h2 = get_md5_hash(self._original_metadata_lookup)
        return h1 != h2

    def __len__(self):
        return len(self.metadata)

    def _get_zarr_group(self, mode="r"):
        return self.workflow._get_workflow_root_group(mode).event_log

    @property
    def metadata(self):
        return self._get_zarr_group().metadata

    @property
    def content(self):
        return self._get_zarr_group().content

    @classmethod
    def generate_in(cls, zarr_group):
        """Generate the persistent data to represent a new event log in a new workflow."""
        event_log = zarr_group.create_group("event_log")
        event_log.attrs["metadata_lookup"] = {
            "event_type": [],
            "machine": [],
        }
        event_log.empty(
            name="metadata",
            dtype=cls._metadata_dtype,
            shape=0,
        )
        event_log.empty(
            name="content",
            dtype=object,
            shape=0,
            object_codec=numcodecs.MsgPack(),
        )

    def _update_metadata_lookup(self, key, item):

        if item in self._metadata_lookup[key]:
            return self._metadata_lookup[key].index(item)
        else:
            self._metadata_lookup[key].append(item)
            return len(self._metadata_lookup[key]) - 1

    @property
    def num_saved_events(self):
        return len(self.metadata)

    @property
    def num_events(self):
        return self.num_saved_events + len(self._new_event_metadata)

    @property
    def has_unsaved_events(self):
        return len(self._new_event_metadata) > 0

    @staticmethod
    def get_timestamp():
        return datetime.utcnow()

    def dump_new_events(self):
        """"""
        if not self._new_event_metadata:
            # no changes
            return

        zarr_group = self._get_zarr_group(mode="r+")

        if self._check_metadata_lookup_changed():
            zarr_group.attrs.put({"metadata_lookup": self._metadata_lookup})
            self._set_original_metadata_lookup()

        zarr_group.metadata.append(np.concatenate(self._new_event_metadata))
        zarr_group.content.append(self._new_event_content)

        self._new_event_metadata = []
        self._new_event_content = []

    def discard_new_events(self):

        self._new_event_metadata = []
        self._new_event_content = []

        if self._check_metadata_lookup_changed():
            self._metadata_lookup = copy.deepcopy(self._original_metadata_lookup)

    def _save_new_events(self):
        if not self.workflow._in_batch_mode:
            self.dump_new_events()

    def _add_event(
        self,
        event_type,
        timestamp=None,
        machine=None,
        parents=None,
        content=None,
    ):

        # TODO: log to app log.

        evt_time = timestamp or self.get_timestamp()

        evt_machine = machine or self.workflow.app.config.get("machine")
        evt_machine_idx = self._update_metadata_lookup("machine", evt_machine)
        evt_type_idx = self._update_metadata_lookup("event_type", event_type)
        evt_idx = self.num_events

        evt_depth = len(parents) if parents else 0
        evt_parent = parents[-1] if parents else -1

        new_content = content or {}
        new_metadata = np.array(
            [
                (
                    evt_idx,
                    evt_time,
                    evt_type_idx,
                    evt_machine_idx,
                    evt_depth,
                    evt_parent,
                )
            ],
            dtype=self._metadata_dtype,
        )

        self._new_event_metadata.append(new_metadata)
        self._new_event_content.append(new_content)

        self._save_new_events()
        evt = self._construct_event(evt_content=new_content, evt_meta=new_metadata[0])

        return evt

    def event_workflow_create(self, timestamp, machine):
        return self._add_event(
            event_type="create_workflow",
            timestamp=timestamp,
            machine=machine,
        )

    def event_add_task(self, new_task_index, **kwargs):
        return self._add_event(
            event_type="add_task",
            content={"new_task_index": new_task_index},
            **kwargs,
        )

    def event_add_empty_task(
        self, new_task_index, added_shared_data, new_parameter_groups, **kwargs
    ):
        return self._add_event(
            event_type="add_empty_task",
            content={
                "new_task_index": new_task_index,
                "added_shared_data": added_shared_data,
                "new_parameter_groups": new_parameter_groups,
            },
            **kwargs,
        )

    def event_add_elements(self, task_index, **kwargs):
        return self._add_event(
            event_type="add_elements",
            content={
                "task_index": task_index,
            },
            **kwargs,
        )

    def event_add_element_set(
        self, task_index, new_element_indices, new_parameter_groups, **kwargs
    ):
        return self._add_event(
            event_type="add_element_set",
            content={
                "task_index": task_index,
                "new_element_indices": new_element_indices,
                "new_parameter_groups": new_parameter_groups,
            },
            **kwargs,
        )

    def event_remove_task(self, task_index, reason, **kwargs):
        return self._add_event(
            event_type="remove_task",
            content={
                "task_index": task_index,
                "reason": reason,
            },
            **kwargs,
        )

    def _construct_event(self, evt_content, evt_meta, meta_fields=None, as_json=False):
        meta_fields = meta_fields or self.metadata.dtype.fields.keys()
        evt = {}
        for i, j in zip(meta_fields, evt_meta):
            if i in self._metadata_lookup:
                j = self._metadata_lookup[i][j]
            evt[i] = j
        evt.update(content=evt_content)
        if not as_json:
            evt = EventLogItem(**evt, event_log=self)
        return evt

    def get_events(self, event_slice=None, as_json=False):

        if not event_slice:
            event_slice = slice(None)  # get all events

        out = []
        meta_fields = self.metadata.dtype.fields.keys()

        # first get events from the persistent store (using slices on the zarr arrays):
        for evt_content, evt_meta in zip(
            self.content[event_slice], self.metadata[event_slice]
        ):
            out.append(self._construct_event(evt_content, evt_meta, meta_fields, as_json))

        # then add any remaining slice indices from the pending updates:
        for evt_idx in range(self.num_events)[event_slice]:
            if evt_idx in range(self.num_saved_events):
                continue
            evt_content = self._new_event_content[evt_idx - self.num_saved_events]
            evt_meta = self._new_event_metadata[evt_idx - self.num_saved_events][0]
            out.append(self._construct_event(evt_content, evt_meta, meta_fields, as_json))

        return out

    def get_event(self, event_index, as_json=False):
        if event_index < 0:
            event_index = self.num_events + event_index
        return self.get_events(
            event_slice=slice(event_index, event_index + 1), as_json=as_json
        )[0]

    def get_events_of_type(self, event_type, as_json=False):
        type_idx = self._metadata_lookup["event_type"].index(event_type)
        idx = np.where(self.metadata["event_type"] == type_idx)[0]
        return [self.get_event(i, as_json=as_json) for i in idx]

    def format_events(self, event_slice=None, full=False, max_line_length=90):
        events = self.get_events(event_slice)

        num_events = len(self)
        idx_size = len(str(num_events))
        out = []
        for evt in events:
            if not full:
                evt_time = np.datetime_as_string(evt.time, unit="s").replace("T", " ")
                parent = f"[{evt.parent:>{idx_size}}] " if evt.parent is not None else ""
                if parent:
                    parent = (("-" * idx_size) + "--") * evt.depth + parent
                out.append(
                    f"{evt.index:>{idx_size}} {evt_time} ({evt.machine}): "
                    f"{parent}{evt.event_type}"
                )
            else:
                evt_time = np.datetime_as_string(evt.time, unit="us").replace("T", " ")
                out.append(f"index:   {evt.index}")
                out.append(f"parent:  {evt.parent}")
                out.append(f"depth:   {evt.depth}")
                out.append(f"time:    {evt.time}")
                out.append(f"type:    {evt.event_type}")
                out.append(f"machine: {evt.machine}")
                if evt.content:
                    out.append("content:")
                    for k, v in evt.content.items():
                        out.append(f"  {k}: {v}")

                out.append("")

        out_str = ""
        for ln in out:
            if len(ln) > max_line_length:
                # remove ansi escape sequences, since we might truncate them:
                ln = remove_ansi_escape_sequences(ln)
                ln = ln[: max_line_length - 3] + ("...")

            out_str += ln + "\n"

        return out_str
